#!/usr/bin/perl


#       Copyright Joshua M. Mitchell and Hunter N.B. Moseley, 2014. All rights reserved.
#       Written by Joshua Mitchell 5/1/2014


package MolFile;

#Load module to enable accessor auotogeneration

use base 'Class::Accessor::Fast';

#Load module to enable subroutine caching

#Load  custom modules

use Molecule; 
use Atom;
use Bond;

#Load pragma

use strict;
#use warnings;

my @mol_file_fields =
  (
  "File"           => 0,  #a string with a file name / path containing a mol file or mol files
  "Type"           => '',  #a string indicating if the molfile contains functional groups or compounds
  "ID"             => '' #if set will append this ID string to db ids. Not used often . 
  );

#Make accessors for MolFile class

MolFile->mk_accessors( grep { $_ ne "0"; } (@mol_file_fields));

#Constructs new object of class 'MolFile'

sub new
  {
  my $proto = shift @_;
  my $self = { @mol_file_fields, @_ };
  $self = bless ( $self, ref ($proto) || $proto );
  #print shift @_;
  return $self;
  }

#Parse the file specified by the given MolFile class and return a hash of molecule objects for each compound in the file

sub parse
  {
  my $self = shift @_;
  my $option = shift @_;
  if ($self->Type eq 'functional')
    {
    $self->parseFunctional();
    }
  elsif ($option eq 'no_hydrogen')
    {
    $self->parse_noHydrogens();
    }
  elsif ($self->Type eq 'compound')
    {
    $self->parseCompound();
    }
  elsif ($self->Type eq 'kcf')
    {
    $self->parse_newKEGG();
    }
  }


#parse compound takes a molfile object ($self) and parses it. This will only work if it is a KEGG or HMDB molfile. Returns a hash of molecule objects representing each entry.  
sub parseCompound
  {
#  print "1\n";
  my $self = shift @_;
  open FILE, ($self->File) or die "ERROR - UNABLE TO OPEN File: $!";
  my $molecule_hash = {};
  my ($bonds, $atoms, $molecule, $molecule_name, $element, $charge, $atom_1, $atom_2, $bond_order, $count, $atom_index, $bond_index, $ID);
  my $string;
  my $count2 ++;
  while (my $line = <FILE>)
    {
    if ($line =~ /^Begi/)
      {
      $count2 ++;
      $string = '';
      ( ($ID, $molecule_name) = ($line =~ /Beginning of SDF File of HMDB(.....)\s+(.+)$/) ) if ( $line =~ /Beginning of SDF File of HMDB/ );
#      print $ID . "\n";
      ( ($ID, $molecule_name) = ($line =~ /Beginning of SDF file of C(\d\d\d\d\d)\s(.+)$/ ) ) if ( $line =~ /Beginning of SDF file of C\d/ );
      $molecule_name =~ s/\s+$//;
#      print "$molecule_name\n";
      if ( $molecule_name eq "" ) {  $molecule_name = "Unknown.$count"; }
      ($bonds, $atoms) = ( [],[] );
      $string = $string . $line;
      }
    elsif ($line =~ /^\s{3,}\-?\d+\.\d+\s+/)
      {
      ($element, $charge) = ($line =~ /^\s{3,}\-?\d+\.\d+\s+\-?\d+\.\d+\s+\-?\d+\.\d+\s{1}(\b.{1,2}\b)\s+\d\s+(\d)/);
      my $value = $#{$atoms} + 2;
      push @$atoms, Atom->new( "element" => $element, "SDF_charge" => $charge, "index" => $#{$atoms} + 2);
      
      $string = $string . $line;
      }
    elsif ($line =~ /^(\s+)?\d+(\s+)?\d+\s+\d+\s+/ and $line !~ /V2000/)
      {
      $string = $string . $line;
      ($atom_1, $atom_2, $bond_order) = ($line =~ /^(...)(...)(...)/);
      $atom_1 =~ s/\s+//;
      $atom_2 =~ s/\s+//;
      $bond_order =~ s/\s+//;
      push @$bonds, Bond->new( "Atom_1" => $atom_1, "Atom_2" => $atom_2, "Bond_order" => $bond_order);
      }
    elsif ($line =~ /M\s\sEND/)
      {
      $string = $string . $line;
      $ID = $self->ID . $ID;
      $$molecule_hash{$molecule_name} = Molecule->new( "Name" => $molecule_name, "Atoms" => $atoms, "Bonds" => $bonds, "ID" => $ID, "String" => $string);
      }
    else
      {    
      $string = $string .$line;
      }
    }
  close FILE;
  return $molecule_hash;
  }
  
#parse compound takes text ($self) and parses it. This will only work if it is a KEGG or HMDB molfile in raw text format being fed in as an array ref ($self). Returns a hash of molecule objects representing each entry.    
sub parseCompound2
  {
#  print "1\n";
  my $self = shift @_;
#  open FILE, ($self->File) or die "ERROR - UNABLE TO OPEN File: $!";
  my $molecule_hash = {};
  my ($bonds, $atoms, $molecule, $molecule_name, $element, $charge, $atom_1, $atom_2, $bond_order, $count, $atom_index, $bond_index, $ID);
  my $string;
  while ( @$self )
    {
    my $line = shift @$self;
    if ($line =~ /^Begi/)
      {
      $count ++;
      $string = '';
      ( ($ID, $molecule_name) = ($line =~ /Beginning of SDF File of HMDB(.....)\s+(.+)$/) ) if ( $line =~ /Beginning of SDF File of HMDB/ );
#      print $ID . "\n";
      ( ($ID, $molecule_name) = ($line =~ /Beginning of SDF file of C(\d\d\d\d\d)\s(.+)$/ ) ) if ( $line =~ /Beginning of SDF file of C\d/ );
      $molecule_name =~ s/\s+$//;
#      print "$molecule_name\n";
      if ( $molecule_name eq "" ) {  $molecule_name = "Unknown.$count"; }
      ($bonds, $atoms) = ( [],[] );
      $string = $string . $line;
      }
    elsif ($line =~ /^\s{3,}\-?\d+\.\d+\s+/)
      {
      ($element, $charge) = ($line =~ /^\s{3,}\-?\d+\.\d+\s+\-?\d+\.\d+\s+\-?\d+\.\d+\s{1}(\b.{1,2}\b)\s+\d\s+(\d)/);
      my $value = $#{$atoms} + 2;
      push @$atoms, Atom->new( "element" => $element, "SDF_charge" => $charge, "index" => $#{$atoms} + 2);
      
      $string = $string . $line;
      }
    elsif ($line =~ /^(\s+)?\d+(\s+)?\d+\s+\d+\s+/ and $line !~ /V2000/)
      {
      $string = $string . $line;
      ($atom_1, $atom_2, $bond_order) = ($line =~ /^(...)(...)(...)/);
      $atom_1 =~ s/\s+//;
      $atom_2 =~ s/\s+//;
      $bond_order =~ s/\s+//;
      push @$bonds, Bond->new( "Atom_1" => $atom_1, "Atom_2" => $atom_2, "Bond_order" => $bond_order);
      }
    elsif ($line =~ /M\s\sEND/)
      {
      $string = $string . $line;
#      $ID = $self->ID . $ID;
      $$molecule_hash{$molecule_name} = Molecule->new( "Name" => $molecule_name, "Atoms" => $atoms, "Bonds" => $bonds, "ID" => $ID, "String" => $string);
      }
    else
      {    
      $string = $string .$line;
      }
    }
#  close FILE;
  return ($molecule_hash);
  }  
  
  
#this takes a molfile object of functional groups ($self) and parses it. Returns a hash of molecule objects representing each entry. 
sub parseFunctional
  {
#  print "2\n";
  my $self = shift @_;
  open FILE, ($self->File) or die "ERROR- UNABLE TO OPEN File: $!";
  my $molecule_hash = {};
  my ($string, $bonds, $atoms, $molecule, $functional_group_name, $id, $element, $charge, $atom_1, $atom_2, $bond_order, $count, $atom_index, $bond_index, $context_atoms);
  while (my $line = <FILE>)
    {
    $string .= $line;
    if ($line =~ /^Begi/)
      {
#      print $line;
      ($id, $functional_group_name) = ($line =~ /Beginning of SDF File of (FGroup\d+)\s+(.+)$/);
      $functional_group_name =~ s/\s+$//;
#      print $functional_group_name, "\n";
      ($bonds, $atoms, $context_atoms) = ( [], [], [] );

      }
    elsif ($line =~ /^\s{3,}\-?\d+\.\d+\s+/)
      {
#      print $line;
      ($element, $charge) = ($line =~ /^\s{3,}\-?\d+\.\d+\s+\-?\d+\.\d+\s+\-?\d+\.\d+\s{1}(!?\b.*?\b)\s+\d\s+(\d)/);;
#      print $element, "\n";
      push @$atoms, Atom->new( "element" => $element, "SDF_charge" => $charge, "index" => $#{$atoms} + 2); 
#      push @$context_atoms, $#{$atoms} if ! ( $line =~ /\*/ );
      push @$context_atoms, $#{$atoms} if ( $line =~/\*/ );
#      print join(",",@$context_atoms) . "\n";
      }
    elsif ($line =~ /^\s+\d+\s+\d+\s+\d+\s+/ and $line !~ /V2000/)
      {
#      print $line;
          ($atom_1, $atom_2, $bond_order) = ($line =~ /^\s+(\d{1,3})\s+(\d{1,3})\s+(\d+)/);
      push @$bonds, Bond->new( "Atom_1" => $atom_1, "Atom_2" => $atom_2, "Bond_order" => $bond_order);
      }
    elsif ($line =~ /M\s\sEND/ or $line=~ /\s\sEND/)
      {
#      print $line;
      $$molecule_hash{$functional_group_name} = Molecule->new( "ID" => $id, "String" => $string, "Context" => $context_atoms, "Name" => $functional_group_name, "Atoms" => $atoms, "Bonds" => $bonds);
      $string = '';
      }
    else
      {
      }
    }
  close FILE;
  return $molecule_hash;
  }

#same as parse compound but will skip hydrogen entries making the appropriate adjustments to indices. Not used in any program currently 
sub parse_noHydrogens
  {
#  print "3\n";
  my $self = shift @_;
  open FILE, ($self->File) or die "ERROR - UNABLE TO OPEN File: $!";
  my $molecule_hash = {};
  my ($bonds, $atoms, $molecule, @hydrogen_indexes, $molecule_name, $element, $charge, $atom_1, $atom_2, $bond_order, $count, $atom_index, $bond_index, $ID);
  while (my $line = <FILE>)
    {
    if ($line =~ /^Begi/)
      {
#      print $line . "\n";
      @hydrogen_indexes = undef;
      $count = 0;
      ( ($ID, $molecule_name) = ($line =~ /Beginning of SDF File of HMDB(.....)\s+(.+)$/) ) if ( $line =~ /Beginning of SDF File of HMDB/ );
      ( ($ID, $molecule_name) = ($line =~ /Beginning of SDF file of C(\d\d\d\d\d)\s(.+)$/ ) ) if ( $line =~ /Beginning of SDF file of C\d/ );
      $molecule_name =~ s/\s+$//;
#      print $molecule_name . "\n";
      if ( $molecule_name eq "") { $molecule_name = "Unknown.$count"; }
      ($bonds, $atoms) = ([],[]);
      }
    elsif ($line =~ /^\s{3,}\-?\d+\.\d+\s+/)

      {
  
        ($element, $charge) = ($line =~ /^\s{3,}\-?\d+\.\d+\s+\-?\d+\.\d+\s+\-?\d+\.\d+\s{1}(\b.{1,2}\b)\s+\d\s+(\d)/);
      $count ++;
      if ( $element =~ /H$/ and $element !~ /!H/ )
        {
        push @hydrogen_indexes, $count;
        next;
        } 
      push @$atoms, Atom->new( "element" => $element, "SDF_charge" => $charge, "index" => $#{$atoms} + 2);
      }
    elsif ($line =~ /^(\s+)?\d+(\s+)?\d+\s+\d+\s+/ and $line !~ /V2000/ and $line !~ /SDF File/)
      {
      print $line . "\n";
      ($atom_1, $atom_2, $bond_order) = ($line =~ /^(...)(...)(...)/);
      $atom_1 =~ s/\s+//;
      $atom_2 =~ s/\s+//;
      $bond_order =~ s/\s+//;
      if ( grep { ( $atom_1 == $_ ) or ( $atom_2 == $_ ) } @hydrogen_indexes )
        {
        next;
        }
      my $remove = undef;
      my ( $remove_1, $remove_2 ) = ( ( grep { $atom_1 > $_ } (@hydrogen_indexes )) - 1 , ( grep { $atom_2 > $_ } (@hydrogen_indexes ) ) - 1 );
      ( $atom_1, $atom_2 ) = ( $atom_1 - $remove_1, $atom_2 - $remove_2 );
      push @$bonds, Bond->new( "Atom_1" => $atom_1, "Atom_2" => $atom_2, "Bond_order" => $bond_order);
      }
    elsif ($line =~ /^M\s+END/)
      {
      if ( @$atoms > 0 )
        {
        $$molecule_hash{$molecule_name} = Molecule->new( "Name" => $molecule_name, "Atoms" => $atoms, "Bonds" => $bonds, "ID" => $ID);
        }
      }
    else
      {

      }
    }
  close FILE;
  return $molecule_hash;
  }

#parse newKEGG is for .kcf as parsecompound is for .mol. take a molfile object with kcf files and it parses it, returning a hash of molecule objects for each entry.   
sub parse_newKEGG
  {
  my $flag = 0;
  my $self = shift @_;
  my $molecule_hash = {};
  my ($bonds, $atoms, $molecule, $molecule_name, $element, $charge, $atom_1, $atom_2, $bond_order, $count, $atom_index, $bond_index, $ID);
  my $string;
  my $duplicates;
  my $ignore;
  open FILE, ($self->File);
  while ( my $line = <FILE> )
    {
    if ($line =~ /^n/)
      {
      $line = substr $line, 1; 
      }
#    print $line;
    if ($line =~ /^ENTRY/ and $flag == 0)
      {
      $string =  $line;
      $count ++;
      ($ID) = ($line =~ /^ENTRY\s*(C.....)/);
      ($bonds, $atoms) = ([],[]);
#      print $ID . "\n";
      }
    elsif ($line =~ /\/\/\//)
      {
      $string = $string . $line;
      $flag = 0;
      if (exists $$molecule_hash{$molecule_name} or $ignore == 1)
        {
        $string = '';
        $atoms = [];
        $bonds = [];
        $molecule_name = undef;
        $ignore = 0;
        next;
        }
      if (@$atoms>0)
        {
        $$molecule_hash{$molecule_name} = Molecule->new( "Name" => $molecule_name, "Atoms" => $atoms, "Bonds" => $bonds, "ID" => $ID, "String" => $string );
        }
      }
    elsif ($line =~ /NAME/)
      {
      $string = $string . $line;
      ($molecule_name) = ($line =~ /NAME\s+(.*);?/);
#      print $molecule_name . "\n";     
      }  
    elsif ($line =~ /ATOM/)
      {
      $string = $string . $line;
#      print $line;
      $flag = 1;
      }
    elsif ($line =~ /BOND/)
      {
      $string = $string . $line;
      $flag = 2;
      }
    elsif ($line =~ /BRACKET/)
      {
      $string = $string . $line;
      $flag = 3;
      }
    elsif ($line !~ /BOND/ and $flag == 1)
      {
      $string = $string . $line;
#      print $line;
      ($element, $charge) = ($line =~ /\s*\d+\s+\w+\s+(.*?)\s+\-?\d+\.\d+\s*\-?\d+\.\d+\s*#?(.*)/);
      push @$atoms, Atom->new( "element" => $element, "SDF_charge", => $charge, "index" => $#{$atoms} + 2);
#      print $element, $charge, "\n";
       if ($element eq '*')
        {
        $ignore = 1;
        }
      }
    elsif ($flag == 2)
      {
      $string = $string . $line;
      ($atom_1, $atom_2, $bond_order) = ($line =~ /\s*\d+\s*(\d+)\s*(\d+)\s*(\d+)/);
      push @$bonds, Bond->new( "Atom_1" => $atom_1, "Atom_2" => $atom_2, "Bond_order" => $bond_order);
#      print $atom_1, $atom_2, $bond_order, "\n";
      }
    }
  close FILE;
  return ($molecule_hash);
  }

#this takes text in an array ref ($self) and parses it as if it is .kcf files. Returns a hash of molecule objects representing each entry. 

sub parse_newKEGG2
  {
  my $flag = 0;
  my $self = shift @_;
  my $molecule_hash = {};
  my ($bonds, $atoms, $molecule, $molecule_name, $element, $charge, $atom_1, $atom_2, $bond_order, $count, $atom_index, $bond_index, $ID);
  my $string;
  my $duplicates;
  my $ignore;
#  open FILE, ($self->File);
  while ( @$self )
    {
    $count ++;
    my $line = shift @$self;
    if ($line =~ /^n/)
      {
      $line = substr $line, 1; 
      }
#    print $line;
    if ($line =~ /^ENTRY/ and $flag == 0)
      {
      $count ++;
      ($ID) = ($line =~ /^ENTRY\s*(C.....)/);
      ($bonds, $atoms) = ([],[]);
#      print $ID . "\n";
      }
    elsif ($line =~ /\/\/\//)
      {
      $flag = 0;
      if (exists $$molecule_hash{$molecule_name} or $ignore == 1)
        {
        $atoms = [];
        $bonds = [];
        $molecule_name = undef;
        $ignore = 0;
        next;
        }
      if (@$atoms>0)
        {
        $$molecule_hash{$molecule_name} = Molecule->new( "Name" => $molecule_name, "Atoms" => $atoms, "Bonds" => $bonds, "ID" => $ID );
        }
      }
    elsif ($line =~ /NAME/)
      {
      ($molecule_name) = ($line =~ /NAME\s+(.*);?/);
      if ($molecule_name eq '')
        {
        $molecule_name = "UNKNOWN";
        }
#      print $molecule_name . "\n";     
      }  
    elsif ($line =~ /ATOM/)
      {
#      print $line;
      $flag = 1;
      }
    elsif ($line =~ /BOND/)
      {
      $flag = 2;
      }
    elsif ($line =~ /BRACKET/)
      {
      $flag = 3;
      }
    elsif ($line !~ /BOND/ and $flag == 1)
      {
#      print $line;
      ($element, $charge) = ($line =~ /\s*\d+\s+\w+\s+(.*?)\s+\-?\d+\.\d+\s*\-?\d+\.\d+\s*#?(.*)/);
      push @$atoms, Atom->new( "element" => $element, "SDF_charge", => $charge, "index" => $#{$atoms} + 2);
#      print $element, $charge, "\n";
       if ($element eq '*')
        {
        $ignore = 1;
        }
      }
    elsif ($flag == 2)
      {
      ($atom_1, $atom_2, $bond_order) = ($line =~ /\s*\d+\s*(\d+)\s*(\d+)\s*(\d+)/);
      push @$bonds, Bond->new( "Atom_1" => $atom_1, "Atom_2" => $atom_2, "Bond_order" => $bond_order);
#      print $atom_1, $atom_2, $bond_order, "\n";
      }
    }
  close FILE;
  return ($molecule_hash);
  }




1;
