#!/usr/bin/perl


#       Copyright Joshua M. Mitchell and Hunter N.B. Moseley, 2014. All rights reserved.
#       Written by Joshua Mitchell 5/1/2014


package Molecule;

use base 'Class::Accessor::Fast';
use Atom;
use Bond;
use Matrix;
use strict;

my %valence_lists = 
            (
            "H" => [1], "Li" => [1],
            "Be" => [2], "B" => [3],
            "C" => [4], "N" => [3,5],
            "O" => [2], "F" => [1],
            "Na" => [1], "Ca" => [2],
            "P" => [3,5], "S" => [2,4,6],
            "I" => [1], "W" => [4,6],
            "K" => [1], "Li" => [1], 
            "Be" => [2], "B" => [3], 
            "F" => [1], "Mg" => [2], 
            "Al" => [3], "Si" => [4], 
            "Sc" => [3], "Ti" => [4], 
            "V" => [5], "Cr" => [6], 
            "Mn" => [2,4,7], "Fe" => [2,3], 
            "Co" => [2,3], "Ni" => [2],
            "Cu" => [2], "Zn" => [2], 
            "Ga" => [3], "Ge" => [2,4], 
            "As" => [3,5], "Se" => [2,4,6], 
            "Br" => [1,3,5,7], "Kr" => [2], 
            "Rb" => [1], "Sr" => [2], 
            "Y" => [3], "Zr" => [4], 
            "Nb" => [5], "Mo" => [4,6], 
            "Tc" => [4,7], "Ru" => [3,4],
            "Rh" => [3], "Pd" => [2,4], 
            "Ag" => [1], "Cd" => [2], 
            "In" => [3], "Sn" => [2,4], 
            "Sb" => [3.5], "Te" => [2,4,6], 
            "I" => [1,3,5,7], "Xe" => [2,4,6,8], 
            "Cs" => [1], "Ba" => [2], 
            "La" => [3], "Ce" => [3,4], 
            "Pr" => [3], "Nd" => [3], 
            "Pm" => [3], "Sm" => [3], 
            "Eu" => [2,3], "Gd" => [3], 
            "Tb" => [3], "Dy" => [3], 
            "Ho" => [3], "Er" => [3], 
            "Tm" =>[3], "Yb" => [3], 
            "Lu" => [3], "Hf" => [4], 
            "Ta" => [5], "W" => [4,6],            
            "Re" => [4], "Os" => [4],
            "Ir" => [3,4], "Pt" => [2,4],
            "Au" => [3], "Hg" => [1,2],
            "Tl" => [1], "Pb" => [2,4],
            "Bi" => [3], "Po" => [2,4],
            "At" => [1], "Rn" => [2],
            "Fr" => [1], "Ra" => [2],
            "Ac" => [3], "Th" => [4],
            "Pa" => [5], "U" => [6], 
            "Np" => [5], "Pu" => [4], 
            "Am" => [3], "Cm" => [3],
            );

my @molecule_fields =
  (
  "Name"              => 0,
  "Atoms"             => 0,
  "Bonds"             => 0,
  "AdjMatrix"         => 0,
  "Formula"           => 0,
  "ExtendedFormula"   => 0,
  "ID"                => 0,
  "part_of"           => 0,
  "String"            => 0,
  "Functionals"      => 0,
  "Groups"             => 0,
  "Context"           => 0,
  "MolFile"           => 0
  );

Molecule->mk_accessors( grep { $_ ne "0"; } (@molecule_fields) );

sub new
  {
  my $proto = shift @_;
  my $self = { @molecule_fields, @_ };
  $self = bless ($self, ref($proto) || $proto);
  $self->countBonds();
  $self->getFormula();
  ($$self{Functionals},$$self{FCopy}) = ([],[]);
  return $self;
  }

#color takes the $self molecule object and colors each node according to what it is bonded to N vertices away. 
#$iterate is the value of N and it will color according to 1 vertex away, 2 vertex away and so on until it gets to N.
##clear is an option, if defined it deletes current color before coloring again. Not many uses for this
#color is good for testing connectivity but becomes slow for high N.  

sub color
  {
  my ($self, $iterate, $clear) = @_;
  my $first = 1;
  my @bonds = @{$$self{Bonds}};
  my @atoms = @{$$self{Atoms}};
  while ($iterate)
    {
    if ($first and $clear)
      {
      foreach my $atom (@atoms)
        {
        $$atom{Super_color} = '';
        $$atom{Color} = '';
        }
      }
    if ($first)
      {
      foreach my $bond (@bonds)
        {
        my $atom1 = $atoms[$$bond{Atom_1}-1];
        my $atom2 = $atoms[$$bond{Atom_2}-1];
        $$atom1{Super_color} = $$atom1{Super_color} . "(" . $$atom2{element} . $$bond{Bond_order} . ");";
        $$atom2{Super_color} = $$atom2{Super_color} . "(" . $$atom1{element} . $$bond{Bond_order} . ");";
        }
      foreach my $atom (@atoms)
        {
        $$atom{Color} = join(";",sort(split(";",$$atom{Super_color},$$atom{Color})));
        }
      }
    else
      {
      foreach my $bond (@bonds)
        {
        my $atom1 = $atoms[$$bond{Atom_1}-1];
        my $atom2 = $atoms[$$bond{Atom_2}-1];
        $$atom1{Super_color} = $$atom1{Super_color} . $$atom2{Color} . ";";
        $$atom2{Super_color} = $$atom2{Super_color} . $$atom1{Color} . ";";
        }
      foreach my $atom (@atoms)
        {
        $$atom{Color} = join(";",sort(split(";",$$atom{Super_color},$$atom{Color})));
        }
      }
      $first = 0;
      $iterate --;
    }
  }

#totalcolor takes all the colors of the nodes and sorts them into a single string that is the molecules color. $self is molecule object. 
sub totalcolor
  {
  my $self = shift @_;
  my $string;
  my $atoms = $$self{Atoms};
  foreach my $atom (@$atoms)
    {
    $string = $string . $$atom{Color} . "!!";
    }
  $string = join("!!",sort(split("!!", $string)));
  return $string;
  }

#countbonds assings bond count and bond order count to all atom objects in the $self molecule object.
sub countBonds
  {
  my $self = shift @_;
  my @bond_count; 
  my @bond_order_count;
  map ( { $bond_count[$$_{Atom_1}-1]+=1 and $bond_count[$$_{Atom_2}-1]+=1 and $bond_order_count[$$_{Atom_1}-1] += $$_{Bond_order} and $bond_order_count[$$_{Atom_2}-1] += $$_{Bond_order} ; } (@{$$self{Bonds}} ) );
  map ( { ${$$self{Atoms}}[$_]->bond_order_count($bond_order_count[$_]) and ${$$self{Atoms}}[$_]->bond_count($bond_count[$_]) } (0..$#{$$self{Atoms}}) ); 
  }

#addHydrogens adds implicit hydrogens explicitly to the $self molecule object, creating a new atom object and bond object for each hydrogen.
#countbonds MUST BE RAN FIRST but since new uses countbonds this isn't a real problem. 
sub addHydrogens
  {
  my $self = shift @_;
  my $count = 0;
  my $hydrogen_index = $#{$self->Atoms} + 2;
  my @starting_atoms = @{$self->Atoms};
  foreach my $atom ( @starting_atoms )
    {
    my @valences = grep { ( $$atom{bond_order_count} + $$atom{oxidation_modifier} ) <= $_; } (@{$valence_lists{$$atom{element}}});
    map ( { ( push @{$$self{Bonds}}, Bond->new ("Atom_1" => $$atom{index}, "Atom_2" => $hydrogen_index, "Bond_order" => 1) ) and 
            ( push @{$$self{Atoms}}, Atom->new ("element" => 'H', "index" => $hydrogen_index, "Bond_order" => 1) ) and
            ( $hydrogen_index ++ ) ; } ( 1..($valences[0] - ($$atom{bond_order_count} + $$atom{oxidation_modifier}) ) ) );

    }
  $self->countBonds();
  $self->getFormula();
  }

#this returns the formula for a molecule object, it is split into getFormula and makeFormula for future use with the memoize module or caching
sub getFormula
  {
  my $self = shift @_;
  my $list_of_atoms = &getElementList($self);
  my @list_of_atoms = (sort {$a cmp $b} (@$list_of_atoms));
  $self->Formula(&makeFormula(@list_of_atoms));
  }

sub makeFormula
  {
  my %counter;
  map { $counter{$_}++; } ( @_ );
  my $formula = join("", map { $_ . $counter{$_}; } (sort {$a cmp $b;} (sort keys %counter)));
  return $formula;
  }
    
#for the $self molecule object, it creates an adjacency matrix representing it's connected structure. The output is a matrixobject saved in the molecule object. 
#If you want hydrogens in it, be sure to add hydrogens first. 
sub makeAdjacencyMatrix
  {
  my $self = shift @_;
  my $adjacency_matrix = [ map { [map {0;} (0..$#{$$self{Atoms}}) ] } (0..$#{$$self{Atoms}}) ];
  map ( { $$adjacency_matrix[$$_{Atom_1} - 1][$$_{Atom_2} - 1] = $$_{Bond_order} and $$adjacency_matrix[$$_{Atom_2} -1][$$_{Atom_1}-1] = $$_{Bond_order}; } @{$$self{Bonds}} );
  my $adjmatrix = Matrix->new("Matrix"=>$adjacency_matrix);
  $self->AdjMatrix($adjmatrix);
  } 

#makes a crude mapping matrix based solely on comparison of element types, $self is molecule object, $other is another molecule object to which you are mapping, output is a matrix object
sub makeMappingMatrix
  {
  my ($self, $other) = ($_[0], $_[1]);
  if ( $#{$$self{Atoms}} < $#{$$other{Atoms}} )
    {
    ($self, $other) = ($other, $self);
    }       
   return Matrix-> new( "Matrix" => [ map ( { my $atom = $_; [ map ( { ( &advancedCompare($_, $$atom{element}) ) } @{&getElementList($self)} ) ] ; } @{$$other{Atoms}} ) ] );
  } 

#makes a mapping matrix such as makeMappingMatrix but can be given a 1d array of matched vertices which will be ommitted from comparison.
sub makeMappingMatrix3
  {
  my ($self, $other, $matched) = ($_[0], $_[1], $_[2]);
  if ( $#{$self->Atoms} < $#{$other->Atoms} )
    {
    ($self, $other) = ($other, $self);
    }    
   return Matrix-> new( "Matrix" => [ map ( { my $atom = $_; [ map ( { my $atom_s = $_;  ( &advancedCompare($atom_s->element, $atom->element) * !(grep{ $_ == $atom_s->index;} @$matched ) ) } @{$self->Atoms} ) ] ; } @{$other->Atoms} ) ] );
  }

#same as makeMappingMatrix essentially but only allows mappings between atoms with the same bond counts and bond order counts, useful for stereoisomer testing.  
sub makeMappingMatrix2
  {
  my ( $self, $other ) = ($_[0], $_[1]);
  if ( $#{$self->Atoms} < $#{$other->Atoms} )
    {
    ( $self, $other ) = ( $other, $self );
    }
  return Matrix-> new( "Matrix" => [ map ( { my $atom = $_; [ map ( { ( &advancedCompare($_->element, $atom->element) * ($_->bond_count == $atom->bond_count ) * ( $_->bond_order_count == $atom->bond_order_count ) ) } @{$self->Atoms} ) ] ; } @{$other->Atoms} ) ] );
  }

#returns the list of elements in the $self molecule
sub getElementList
  {
  my $self = shift @_;
  return [ map ( { $$_{element} } @{$$self{Atoms}} ) ];
  }

#makes a mapping matrix using color rules, colors must match in order for a valid mapping to exist. 
sub makeMappingMatrixColor
  {
  my ($self, $other) = ($_[0], $_[1]);
  if ( $#{$$self{Atoms}} < $#{$$other{Atoms}} )
    {
    ($self, $other) = ($other, $self);
    }
  return Matrix->new("Matrix" => [ map ( { my $atom = $_; [ map ( { my $atom_s = $_; 1 * ( &stringCompare($_->Superuber_color,$atom->Superuber_color) ) } @{$self->Atoms} ) ] ; } @{$other->Atoms} ) ] );
  }

#for commonly compared strings such as colors, given $string1 and $string2, two strings, it performs 'eq' on them and cache's the result in the future. 
my %string_cache;
sub stringCompare
  {
  my ($string_1, $string_2) = @_;
  return $string_cache{$string_1}{$string_2} if (exists $string_cache{$string_1}{$string_2});
  my $value = ($string_1 eq $string_2);
  $string_cache{$string_1}{$string_2} = $value;
  $string_cache{$string_2}{$string_1} = $value;
  return $value;
  }

#used to compare element types of $element_1 and $element_2 which are strings. It does the comparison using advancedCompareActual and caches the result. 
#returns 1 if they match, 0 if they don't according to the comparison rules. 
my %cache;
sub advancedCompare
  {
  my ($element_1, $element_2) = @_;
  return $cache{$element_1}{$element_2} if (exists $cache{$element_1}{$element_2});
  my $return = &advancedCompareActual($element_1,$element_2);
  $cache{$element_1}{$element_2} = $return;
  $cache{$element_2}{$element_1} = $return;
  return $return;
  }

#tests is two element types are equivalent including !x and x|y types. 
sub advancedCompareActual
  {
  my ($element_1, $element_2) = @_;
#  print $element_1, "::", $element_2, "\n";
  if ($element_1 eq $element_2)  #simple string comparison.
    {
    return 1;
    }
  elsif ( ($element_1 =~ /!/ and $element_2 =~ /\|/) || ($element_1 =~ /\|/ and $element_2 =~ /!/) || ($element_1 =~ /!/ and $element_2 =~ /!/) ) #unallowed combos of special types
    {
    return 0;
    }
  elsif (($element_1 =~ /\|/ and $element_1 !~ /\!/) or ($element_2 =~ /\|/ and $element_2 !~ /\!/)   ) #elements 1 and 2 are the same types
    {
#    print "\tDOINGTHIS\n";
#    print ($element_1, "\t", $element_2, "\n");
#    my @element_1 = split ("\|", $element_1);

    my @element_1 = grep ( {$_ ne '|'; } split ( /\|/, $element_1) ); #break special types up into element types
    my @element_2 = grep ( {$_ ne '|'; } split ( /\|/, $element_2) );
#    print join("&", @element_1), "\n";
    my $value;
    foreach my $element (@element_1)
      {
#      print $element;
      $value += grep ( {$_ eq $element} ( @element_2 ) ); #do any match?
      }
    if ( $value > 0 ) #yes, so return 1.
      {
      return 1;
      } 
    if ( $value == 0 ) #no, return 0
      {
      return 0;
      }
    }
  elsif (($element_1 =~ /\|/ and $element_1 =~ /\!/))
    {
    my @element_1 = split( /\|/, $element_1);
    my $value;
    foreach my $element (@element_1) 
      {
      $value = &advancedCompareActual($element,$element_2); #split them up and rerun because it is a combination group, !x|!y
      if ($value == 0)
        {
        return 0;
        }
      }
    return 1;
    }
  elsif (($element_2 =~ /\|/ and $element_2 =~ /\!/))i #same as above but for element 2. 
    {
    my @element_2 = split( /\|/, $element_2);
    my $value;
    foreach my $element (@element_2)
      {
      $value = &advancedCompareActual($element, $element_1);
      if ($value == 0) 
        {
        return 0;
        }
      }
    return 1;
    }
  elsif (($element_2 =~ /\|/ and $element_2 =~ /\!/) and ($element_1 =~ /\|/ and $element_1 =~ /\!/ ))
    {
    return 1 if ($element_2 eq $element_1); #just test if they are the same. !C|!H will not match !H|!C but this shouldn't occur as special types aren't in databases only in fgroups. Additionally this limitation is cited here and readme so you should be aware of it, this will be fixed in the future but isn't a huge concern. 
    return 0;
    }




  else  #elsif ($element_1 =~ /!/ or $element_2 =~ /!/)
    {
    my ($element_1_type, $element_2_type) = map { /^[!]?(.+)$/; } (@_);
    ($element_1 =~ /!/ and $element_2 =~ /!/ ) ? return 0: ( ( $element_1 =~ /!/ || $element_2 =~ /!/ ) and ( ($element_1_type eq $element_2_type ) * 1 == 1 )) ? return 0:( ( $element_1 =~ /!/ || $element_2 =~ /!/ ) and ( ($element_1_type eq $element_2_type ) * 1 == 0 )) ? return 1:
    return ( ( $element_1_type eq $element_2_type ) * 1); #split them up and test !H does not equal !H because that would have very permissive matching where A and B are matched but A != B. 
    }
  return;
  }

1;
