#!/usr/bin/perl

#Name: Ullmann.pm
#Contains the methods for the Ullmann object, an object that represents an instance of the Ullmann algorithm. There are experimental methods 
#at the end of this module that are not currently in use but included for completeness. 


#       Copyright Joshua M. Mitchell and Hunter N.B. Moseley, 2014. All rights reserved.
#       Written by Joshua Mitchell 5/1/2014


package Ullmann;

use base 'Class::Accessor::Fast';
use strict;
use warnings; 
use integer;
#Create prototypical ullmann object

my @ullmann_fields = 
  (
# Add variables of the ullmann object here 
  );


Ullmann->mk_accessors( grep { $_ ne "0"; } (@ullmann_fields) );

#Name: new
#Description: Creates a new ullmann object
#Input: None
#Output: Ullmann Object

sub new
  {
  my $proto = shift @_;
  my $self = { @ullmann_fields, @_ };
  return bless ( $self, ref($proto) || $proto );
  }

#Name: enumerateFull
#Description: Searches for all instances of the given subgraph (or compound) in a graph (or another compound)
#	      then returns the indexes of the matched atoms if the entire subgraph is matched. 
#Options: None
#Input: ( Ullmann Object, Mapping Matrix, Subgraph Compound, Graph Compound )
#Output: ( Indexes of Matched Atoms )

sub enumerateFull
  {
  my ($enumerator, $self, $functional_group, $compound) = @_;
#  my $enumerator = shift @_;
#  my $self = shift @_;
#  my $functional_group = shift @_;
#  my $compound = shift @_;
  my $isomorphs = [];
  my $row = 0;
  my $col_max = $#{${$self->Matrix}[0]};
  my $row_max = $#{$self->Matrix};
  my $used = [map {0;} (0..$col_max)];
  my $mapping = [map {-1;} (0..$row_max)];
  while ($row >= 0)
    {
    do
      { $$mapping[$row]++; }
    while ($$mapping[$row] <= $col_max and (${$self->Matrix}[$row][$$mapping[$row]] == 0 || $$used[$$mapping[$row]] == 1));
    next if ($$mapping[$row] > $col_max);
    if ($row == $row_max)
      {
      push @$isomorphs, [$mapping]  if ( $functional_group->AdjMatrix->equate2($compound->AdjMatrix,$mapping));
      }
    elsif ( $row == 0 or $functional_group->AdjMatrix->equate2($compound->AdjMatrix,[ map { $$mapping[$_]; } (0..$row)]))
      {
      $$used[$$mapping[$row]] = 1;
      $row++;
      }
    }
  continue
    {
    while($row >= 0 && $$mapping[$row] >= $col_max)
      {
      $$mapping[$row] = -1;
      $row--;
      $$used[$$mapping[$row]] = 0;
      }
    }
  return $isomorphs;
  }

#Name: enumerateShort
#Description: Searches for a single instance of the given subgraph (or compound) in a graph (or another compound)
#	      then returns the indexes of the matched atoms if the entire subraph is matched.
#Options: None
#Input: (Ullmann Object, Mapping Matrix, Subgraph Compound, Graph Compound)
#Output: (Indexes of Matched Atoms)

sub enumerateShort
  {
  my ($enumerator, $self, $functional_group, $compound) = @_;
  my ($isomorphs, $row, $col_max, $row_max) = ( [], 0, $#{${$self->Matrix}[0]}, $#{$self->Matrix} );
  my ($used, $mapping) = ( [map {0;} (0..$col_max)], [map {-1;} (0..$row_max)] ); 
  while ($row >= 0)
    {
    do
      { $$mapping[$row]++; }
    while ($$mapping[$row] <= $col_max and (${$self->Matrix}[$row][$$mapping[$row]] == 0 || $$used[$$mapping[$row]] == 1));
    next if ($$mapping[$row] > $col_max);
    if ($row == $row_max)
      {
      return [@$mapping]  if ( $functional_group->AdjMatrix->equate2($compound->AdjMatrix,$mapping));
      }
    elsif ( $row == 0 or $functional_group->AdjMatrix->equate2($compound->AdjMatrix,[ map { $$mapping[$_]; } (0..$row)]))
      {
      $$used[$$mapping[$row]] = 1;
      $row++;
      }
    }
  continue
    {
    while($row >= 0 && $$mapping[$row] >= $col_max)
      {
      $$mapping[$row] = -1;
      $row--;
      $$used[$$mapping[$row]] = 0;
      }
    }

  return $isomorphs;
  }


#Name: enumerate
#Description: Searches for instances of a given subgraph in a graph in one of two modes. In the first
#	      enumerate will find all instances of the subgraph, in the second mode, enumerate will
#	      stop searching after the first instance is found. The indices of matched atoms are 
#	      returned when searching is complete. Similar to enumerateFull and enumerateShort.
#Options: option = 0 or not given- exhaustive searching
#	  option = 1- short circuit, return after first match is found.
#Input: (Ullmann Object, Mapping Matrix, Subgraph Compound, Graph Compound)
#Output: (Indexes of Matched Atoms)

sub enumerate
  {
  my ($enumerator, $self, $functional_group, $compound, $option, $array) = @_;
  my $isomorphs = [];
  my $row = 0;
  my $col_max = $#{${$self->Matrix}[0]};
  my $row_max = $#{$self->Matrix};
  my $used = [map {0;} (0..$col_max)];
  foreach my $value (@$array)
    {
    $$used[$value] = 1;
    }
  my $mapping = [map {-1;} (0..$row_max)];
  while ($row >= 0)
    {
    do
      { $$mapping[$row]++; }
    while ($$mapping[$row] <= $col_max and ($$self{Matrix}[$row][$$mapping[$row]] == 0 || $$used[$$mapping[$row]] == 1));
    next if ($$mapping[$row] > $col_max);
    if ($row == $row_max)
      {
      push @$isomorphs, [@$mapping]  if ( $functional_group->AdjMatrix->equate2($compound->AdjMatrix,$mapping));
#      if ( $option eq 'short' )
#        {
#        return [@$mapping]  if ( $functional_group->AdjMatrix->equate2($compound->AdjMatrix,$mapping));
#        }
      }
    elsif ( $row == 0 or $functional_group->AdjMatrix->equate2($compound->AdjMatrix,[ map { $$mapping[$_]; } (0..$row)]))
      {
      $$used[$$mapping[$row]] = 1;
      $row++;
      }
    }
  continue
    {
    while($row >= 0 && $$mapping[$row] >= $col_max)
      {
      $$mapping[$row] = -1;
      $row--;
      $$used[$$mapping[$row]] = 0 unless grep {$_ == $$mapping[$row];} (@$array);
      }
    }
  return $isomorphs;
  }

################################

# The following are works in progress, some are much faster than the methods above but may be unstable and give unexpected results use at your own risk
# they are undocumented as in no comments, but they are derivations of above. Feel free to ignore as well, they aren't used in any programs yet. 

sub enumerateFaster2
  {
  my ($enumerator, $self, $functional_group, $compound, $option, $array_ref) = @_;
  my @array = @{$array_ref};
  my @isomorphs;
  my $row = 0;
  my $col_max = $#{$$self{Matrix}[0]};
  my $row_max = $#{$$self{Matrix}};
  my $self_matrix = $$self{Matrix};
  my @used = map{0;} (0..$col_max);
  my @mapping = map {-1;} (0..$row_max);
  my $length = $#mapping;
  my $functional_adj = $$functional_group{AdjMatrix}{Matrix};
  my $compound_adj = $$compound{AdjMatrix}{Matrix};
  my @sm_rows;
  my @om_rows;
  my %cache;
  foreach my $row (@$functional_adj) { push @sm_rows, $row;}
  foreach my $row (@$compound_adj) { push @om_rows, $row; }
  if (@array)
    {
    foreach my $value (@array)
      {
      $used[$value] = 1;
      }
    }
  while ($row != -1)
    {
    do
      { $mapping[$row]++; }
    while ($mapping[$row] <= $col_max and ($used[$mapping[$row]] == 1 || $$self_matrix[$row][$mapping[$row]] == 0));      
    next if ($mapping[$row] > $col_max);
    if ($row == $row_max)
      {
      my $value2 = 1;
      OUTER1: for (0..$length)
        {
        my $smrow = $sm_rows[$_];
        my $omrow = $om_rows[$mapping[$_]];
        for (0..$length)
          {
          if ($$smrow[$_] != $$omrow[$mapping[$_]])
            {
            $value2 = 0;
            last OUTER1;
            }
          }
        }
      push @isomorphs, \@mapping if ($value2 == 1);
      }
     elsif ( $row == 0 or Matrix::equate2( $functional_adj, $compound_adj, @mapping, $row ))
#     elsif ( $row == 0 or Matrix::equate4( $functional_adj, $compound_adj, [ map { $mapping[$_]; } (0..$row)], $row ))
      {
      $used[$mapping[$row]] = 1;
      $row++;
      }
    }
  continue
    {
    while(($mapping[$row] >= $col_max) && $row >= 0 )
      {
      $mapping[$row] = -1;
      $row--;
      $used[$mapping[$row]] = 0 unless grep {$_ == $mapping[$row];} (@array);
      }
    }
  return \@isomorphs;
  }

sub enumerateFaster4
  {
  my ($enumerator, $self, $functional_group, $compound, $option, $array_ref) = @_;
  my @array = @{$array_ref};
  my @isomorphs;
  my $row = 0;
  my $col_max = $#{$$self{Matrix}[0]};
  my $row_max = $#{$$self{Matrix}};
  my $self_matrix = $$self{Matrix};
  my @used = map{0;} (0..$col_max);
  my @mapping = map {-1;} (0..$row_max);
  my $length = $#mapping;
  my $functional_adj = $$functional_group{AdjMatrix}{Matrix};
  my $compound_adj = $$compound{AdjMatrix}{Matrix};
  my @sm_rows;
  my @om_rows;
  my %cache;
  foreach my $row (@$functional_adj) { push @sm_rows, $row;}
  foreach my $row (@$compound_adj) { push @om_rows, $row; }
  if (@array)
    {
    foreach my $value (@array)
      {
      $used[$value] = 1;
      }
    }
  while ($row != -1)
    {
    do
      { $mapping[$row]+=1; }
    while ( ($mapping[$row] <= $col_max) && ($used[$mapping[$row]] == 1 || $$self_matrix[$row][$mapping[$row]] == 0) );      
    next if ($mapping[$row] > $col_max);
    if ($row == $row_max)
      {
      my $value2 = 1;
      OUTER1: for (0..$length)
        {
        my $smrow = $sm_rows[$_];
        my $omrow = $om_rows[$mapping[$_]];
        for (0..$length)
          {
          if (($$smrow[$_] != $$omrow[$mapping[$_]]))
            {
            $value2 = 0;
            last OUTER1;
            }
          }
        }
#      return \@mapping if ($option == 1 and $value2 == 1);
      push @isomorphs, \@mapping if ($value2 == 1);
      }
     elsif ( ($row == 0) || (Matrix::equate4( $functional_adj, $compound_adj, \@mapping, $row )))
#     elsif ( $row == 0 or Matrix::equate4( $functional_adj, $compound_adj, [ map { $mapping[$_]; } (0..$row)], $row ))
      {
      $used[$mapping[$row]] = 1;
      $row++;
      }
    }
  continue
    {
    while( ($mapping[$row] >= $col_max) && ($row >= 0) )
      {
      $mapping[$row] = -1;
      $row--;
      $used[$mapping[$row]] = 0 unless grep { $_ == $mapping[$row];} (@array);
      }
    }
  return \@isomorphs;
  }

sub enumerateFaster5
  {
  my ($enumerator, $self, $functional_group, $compound, $option, $array_ref) = @_;
  my @isomorphs;
  my $row = 0;
  my $col_max = $#{$$self{Matrix}[0]};
  my $row_max = $#{$$self{Matrix}};
  my $self_matrix = $$self{Matrix};
  my @used = map{0;} (0..$col_max);
  my @mapping = map {-1;} (0..$row_max);
  my $functional_adj = $$functional_group{AdjMatrix}{Matrix};
  my $compound_adj = $$compound{AdjMatrix}{Matrix};
  my @sm_rows;
  my @om_rows;
  my @mm_rows;
  my @indexs;
  my @indices = (0..$row_max);
  foreach my $index (0..$row_max)
    {
    my @array = (0..$row_max);
    my $array2 = [grep {$_ != $index} (@array)];
    $indexs[$index] = $array2;
    }
  foreach my $row (@$self_matrix) { push @mm_rows, $row;}
  foreach my $row (@$functional_adj) { push @sm_rows, $row;}
  foreach my $row (@$compound_adj) { push @om_rows, $row; }
  my $self_row = $mm_rows[0];
  my $integer = -1;
  while ()
    {
    do
      { $integer ++; }
    while ( ($integer <= $col_max) && ( !$$self_row[$integer] || $used[$integer]) ) ;       
    $mapping[$row] = $integer;
    next if ($integer > $col_max);
    if ($row == $row_max)
      {
      my $value2 = 1;
      OUTER1: for ($row_max)
        {
        my $smrow = $sm_rows[$_];
        my $omrow = $om_rows[$mapping[$_]];
        my $stuff = $indexs[$_];
        for (@$stuff)
          {
          if (($$smrow[$_] != $$omrow[$mapping[$_]]))
            {
            $value2 = 0;
            last OUTER1;
            }
          }
        }
      return [@mapping]; 
#      push @isomorphs, \@mapping if ($value2);
      }
     elsif ( !$row || &Matrix::equate4( $functional_adj, $compound_adj, \@mapping, $row ))
      {
      $used[$integer] = 1;
      $row++;
      $self_row = $mm_rows[$row];
      $integer = $mapping[$row];
      }
    }
  continue
    {
    while( ($integer >= $col_max) )
      {
      $mapping[$row] = -1;
      $row--;
      $integer = $mapping[$row];
      $self_row = $mm_rows[$row];
      $used[$integer] = 0;
      }
    last if ($row == -1);
    }
  return \@isomorphs;
  }

sub enumerateFasterFull
  {
  my ($enumerator, $self, $functional_group, $compound, $option, $array_ref) = @_;
  my @isomorphs;
  my $row = 0;
  my $col_max = $#{$$self{Matrix}[0]};
  my $row_max = $#{$$self{Matrix}};
  my $self_matrix = $$self{Matrix};
  my @used = map{0;} (0..$col_max);
  my @mapping = map {-1;} (0..$row_max);
  my $functional_adj = $$functional_group{AdjMatrix}{Matrix};
  my $compound_adj = $$compound{AdjMatrix}{Matrix};
  my @sm_rows;
  my @om_rows;
  my @mm_rows;
  my @indexs;
  my @indices = (0..$row_max);
  foreach my $index (0..$row_max)
    {
    my @array = (0..$row_max);
    my $array2 = [grep {$_ != $index} (@array)];
    $indexs[$index] = $array2;
    }
  foreach my $row (@$self_matrix) { push @mm_rows, $row;}
  foreach my $row (@$functional_adj) { push @sm_rows, $row;}
  foreach my $row (@$compound_adj) { push @om_rows, $row; }
  my $self_row = $mm_rows[0];
  my $integer = -1;
  my $search_yes = 0;
  while ($row != -1)
    {
    do
      { $integer ++; }
    while ( ($integer <= $col_max) && ( !$$self_row[$integer] ) ) ;       
    $mapping[$row] = $integer;
    next if ($integer > $col_max);
    if ($search_yes)
      {
      my $value2 = 1;
      OUTER1: for ($row_max)
        {
        my $smrow = $sm_rows[$_];
        my $omrow = $om_rows[$mapping[$_]];
        my $stuff = $indexs[$_];
        for (@$stuff)
          {
          if (($$smrow[$_] != $$omrow[$mapping[$_]]))
            {
            $value2 = 0;
            last OUTER1;
            }
          }
        push @isomorphs, \@mapping; 
        }
#      return \@mapping if ($option == 1 and $value2 == 1);
      }
     elsif ( !$row || &Matrix::equate4( $functional_adj, $compound_adj, \@mapping, $row ))
      {
      $used[$integer] = 1;
      $row++;
      $self_row = $mm_rows[$row];
      $integer = $mapping[$row];
      $search_yes = ($row==$row_max);
      }
    }
  continue
    {
    while( ($integer >= $col_max) )
      {
      $mapping[$row] = -1;
      $row--;
      $integer = $mapping[$row];
      $self_row = $mm_rows[$row];
      $search_yes = ($row==$row_max);
      }
    }
  return \@isomorphs;
  }

sub enumerateFasterShort
  {
  my ($enumerator, $self, $functional_group, $compound, $option, $array_ref) = @_;
  my @isomorphs;
  my $row = 0;
  my $col_max = $#{$$self{Matrix}[0]};
  my $row_max = $#{$$self{Matrix}};
  my $self_matrix = $$self{Matrix};
  my @used = map{0;} (0..$col_max);
  my @mapping = map {-1;} (0..$row_max);
  my $functional_adj = $$functional_group{AdjMatrix}{Matrix};
  my $compound_adj = $$compound{AdjMatrix}{Matrix};
  my @sm_rows;
  my @om_rows;
  my @mm_rows;
  my @indices;
  foreach my $index (0..$row_max)
    {
    $indices[$index] = [grep {$_ != $index} (0..$row_max)];
    }
  foreach my $row (@$self_matrix) { push @mm_rows, $row;}
  foreach my $row (@$functional_adj) { push @sm_rows, $row;}
  foreach my $row (@$compound_adj) { push @om_rows, $row; }
  my $self_row = $mm_rows[0];
  my $integer = -1;
  my $search_yes = 0;
  while ($row != -1)
    {
    do
      { $integer ++; }
    while ( ($integer <= $col_max) && ( !$$self_row[$integer] ) ) ;       
    $mapping[$row] = $integer;
    next if ($integer > $col_max);
    my $value = 0;
    if ($search_yes)
      {
      OUTER1: for ($row_max)
        {
        my $smrow = $sm_rows[$_];
        my $omrow = $om_rows[$mapping[$_]];
        my $stuff = $indices[$_];
        for (@$stuff)
          {
          if (($$smrow[$_] != $$omrow[$mapping[$_]]))
            {
            last OUTER1;
            }
          }
        return \@mapping; 
        }
#      return \@mapping if ($option == 1 and $value2 == 1);
      }
    else
      {
      if (!$row)
        {
        $used[$integer] = 1;
        $row++;
        $self_row = $mm_rows[$row];
        $integer = $mapping[$row];
        $search_yes = ($row==$row_max);
        } 
      else
        {
        OUTER2: for ($row)
          {
          my $smrow = $sm_rows[$_];
          my $omrow = $om_rows[$mapping[$_]];
          my $a = $_;
          for ($row)
            {
            next if ($_ == $a);
            if ($$smrow[$_] != $$omrow[$mapping[$_]])
              {
              last OUTER2;
              }
            }
          $used[$integer] = 1;
          $row++;
          $self_row = $mm_rows[$row];
          $integer = $mapping[$row];
          $search_yes = ($row==$row_max);
          }
        }
      }
    }
  continue
    {
    while( ($integer >= $col_max) )
      {
      $mapping[$row] = -1;
      $row--;
      $integer = $mapping[$row];
      $self_row = $mm_rows[$row];
      $search_yes = ($row==$row_max);
      }
    }
  return \@isomorphs;
  }




1;
